@@ -2,8 +2,8 @@
 
 ;; Copyright (C) 2025 Vadim Pavlov
 ;; Author: Vadim Pavlov <https://github.com/Prgebish>
-;; Package-Version: 20260218.1503
-;; Package-Revision: 407e3b345c93
+;; Package-Version: 20260225.801
+;; Package-Revision: faa32ab34866
 ;; Package-Requires: ((emacs "27.1"))
 ;; Keywords: convenience, navigation
 ;; URL: https://github.com/Prgebish/flash
@@ -174,6 +174,13 @@ Set to 0 to always show labels (default)."
   :type 'integer
   :group 'flash)
 
+(defcustom flash-search-folds nil
+  "When non-nil, search inside folded (invisible) regions.
+When nil (default), only visible text is searched.
+Flash-isearch always searches folds regardless of this setting."
+  :type 'boolean
+  :group 'flash)
+
 ;;; State for continue functionality
 
 (defvar flash--last-pattern nil
